import geopy.distance

class Graph:
    def __init__(self, adjacent_list: dict, stations_dict: dict, lines: dict):
        self.adjacent_list = adjacent_list # Dictionary with distance between adjacent nodes.
        self.stations_dict = stations_dict # Dictionary with stations and their coordinates.
        self.lines = lines
        self.VELOCIDAD_MEDIA_TREN = 36
        self.VELOCIDAD_MEDIA_HUMANO = 5
        self.FRECUENCIA = 0.0333
        self.TIEMPO_PARADA_POR_ESTACION = 0.0083


    def get_neighbours(self, node: str) -> list:
        """
        To obtain adjacent nodes of the node passed as a parameter.
        """
        return self.adjacent_list[node]

    def reconstruct_path(self, node_couple: dict, min_f_node: str) -> list:
        """
        To reconstruct a sequential path from a list of adjacencies.
        """
        path = []

        while node_couple[min_f_node] != min_f_node:
            path.append(min_f_node)
            min_f_node = node_couple[min_f_node]
        path.append(min_f_node)
        path.reverse()

        return path


    def get_best_node(self, open_list: list, g: list, destination: str) -> str:
        """
        To obtain the node name with the minimum f value.
        """
        min_f_node = open_list[0]

        for node in open_list:
            if g[node] + self.h(node, destination) < g[min_f_node] + self.h(min_f_node, destination):
                min_f_node = node

        return min_f_node

    def h(self, node: str, destination: str) -> float:
        """
        Estimate distance based on a consistent heuristic by calculating
        the distance between the current node and the destination node,
        and adding a transfer value if the nodes belong to different lines.
        """
        transfer = True
        lines = ['Line 1: red', 'Line 2: blue', 'Line 3: green']
        for line in lines:
            if node in self.lines[line] and destination in self.lines[line]:
                transfer = False

        current_node_coordinates = self.stations_dict[node]['coordinates']
        destination_coordinates = self.stations_dict[destination]['coordinates']
        result = geopy.distance.geodesic(current_node_coordinates, destination_coordinates).km
        if transfer:
            result += 0.5

        print(result)
        return result


    def calculate_travel_time(self, distance: float) -> float:
        """
        Calcular el tiempo de viaje entre estaciones en base a la distancia.
        Se considera la velocidad promedio del tren, la velocidad promedio de una persona,
        la frecuencia de los trenes y el tiempo de parada en cada estación.
        """
        tiempo_tren = distance / self.VELOCIDAD_MEDIA_TREN
        tiempo_caminata = distance / self.VELOCIDAD_MEDIA_HUMANO
        tiempo_espera = self.FRECUENCIA
        tiempo_parada = self.TIEMPO_PARADA_POR_ESTACION

        tiempo_total = tiempo_tren + tiempo_caminata + tiempo_espera + tiempo_parada
        return tiempo_total

    def a_star_search(self, origin, destination) -> list:
        open_list = [origin]
        closed = []

        # Dictionary with the distances between the origin and each node (g).
        g = {}
        g[origin] = 0

        # Dictionary with pairs of adjacent nodes.
        node_couple = {}
        node_couple[origin] = origin

        while len(open_list) > 0:
            # Select the node with the minimum f value.
            min_f_node = self.get_best_node(open_list, g, destination)

            # If the node is at the destination, we have arrived.
            if min_f_node == destination:
                reconstructed_path = self.reconstruct_path(node_couple, min_f_node)
                distance = g[min_f_node]
                travel_time = self.calculate_travel_time(distance)
                print(f"Camino: {reconstructed_path}")
                print(f"Distancia: {distance:.2f} km")
                print(f"Tiempo de viaje: {travel_time:.2f} minutos")
                return reconstructed_path

            # Check for all neighbors if they are already in the closed list or the open_list.
            # Then, verify which path is the best and update the cost.
            # If any visited path is better, move it to the to-be-visited list.
            for (name, distance) in self.get_neighbours(min_f_node):
                if name not in open_list and name not in closed:
                    open_list.append(name)
                    node_couple[name] = min_f_node
                    g[name] = g[min_f_node] + distance
                else:
                    if g[name] > g[min_f_node] + distance:
                        g[name] = g[min_f_node] + distance
                        node_couple[name] = min_f_node

                        if name in closed:
                            closed.remove(name)
                            open_list.append(name)
            open_list.remove(min_f_node)
            closed.append(min_f_node)

        print('The path does not exist.')

        return None
